@@ -16,6 +16,7 @@
  */
 
 #include <stdlib.h>
+#include <stdarg.h>
 #include <stdio.h>
 #include <string.h>
 #include <limits.h>
@@ -75,9 +76,12 @@ static void default_values (struct starter_config *cfg)
 
 	cfg->conn_default.left.addr_family = AF_INET;
 	anyaddr(AF_INET, &cfg->conn_default.left.addr);
+	cfg->conn_default.left.nexttype  = KH_NOTSET;
 	anyaddr(AF_INET, &cfg->conn_default.left.nexthop);
+
 	cfg->conn_default.right.addr_family = AF_INET;
 	anyaddr(AF_INET, &cfg->conn_default.right.addr);
+	cfg->conn_default.right.nexttype = KH_NOTSET;
 	anyaddr(AF_INET, &cfg->conn_default.right.nexthop);
 
 	/* default is to look in DNS */
@@ -91,11 +95,35 @@ static void default_values (struct starter_config *cfg)
 	cfg->ctlbase = clone_str(CTL_FILE, "default base");
 }
 
-#define ERR_FOUND(args...) \
-	{ if (perr && (*perr==NULL)) { \
-		snprintf(_tmp_err, sizeof(_tmp_err)-1, ## args); \
-		*perr = xstrdup(_tmp_err); } \
-	err++; }
+/* format error, and append to string of errors */
+int error_append(char **perr, const char *fmt, ...) 
+{
+    va_list args;
+
+    if(perr) {
+	char *nerr; 
+	int len; 
+	
+	va_start(args, fmt);
+	vsnprintf(_tmp_err, sizeof(_tmp_err)-1, fmt, args);
+	va_end(args);
+	
+	len = 1+ strlen(_tmp_err) + (*perr ? strlen(*perr) : 0);	
+	nerr = xmalloc(len);
+	nerr[0]='\0';
+	if(*perr) strcpy(nerr, *perr); 
+	strcat(nerr, _tmp_err);			
+	
+	if(*perr) free(*perr);
+	*perr = nerr;
+
+	return 1;
+    }
+    return 0;
+}
+
+#define ERR_FOUND(args...) do { err += error_append(&err_str, ##args); } while(0)
+
 
 #define KW_POLICY_FLAG(val,fl) if(conn->options_set[val]) \
         { if(conn->options[val]) \
@@ -281,6 +309,7 @@ static int validate_end(struct starter_conn *conn_st
 			, bool left, err_t *perr)
 {
     err_t er = NULL;
+    char *err_str = NULL;
     const char *leftright=(left ? "left" : "right");
     int err=0;
 
@@ -366,16 +395,21 @@ static int validate_end(struct starter_conn *conn_st
     if(end->strings_set[KSCF_NEXTHOP])
     {
 	char *value = end->strings[KSCF_NEXTHOP];
-	
-	end->nexttype = KH_IPADDR;
 
-	er = ttoaddr(value, 0, AF_INET, &(end->nexthop));
-	if (er) ERR_FOUND("bad addr %snexthop=%s [%s]", leftright, value, er);
+	if(strcasecmp(value, "%defaultroute")==0) {
+	    end->nexttype=KH_DEFAULTROUTE;
+	} else {
+	    er = ttoaddr(value, 0, AF_INET, &(end->nexthop));
+	    if (er) ERR_FOUND("bad addr %snexthop=%s [%s]", leftright, value, er);
 
+	    end->nexttype = KH_IPADDR;
+	}
     } else {
+#if 0
 	if(conn_st->policy & POLICY_OPPO) {
 	    end->nexttype = KH_DEFAULTROUTE;
 	}
+#endif
 	anyaddr(AF_INET, &end->nexthop);
     }
 
@@ -467,6 +501,7 @@ static int validate_end(struct starter_conn *conn_st
     KSCF_MAX          = 19
 */
 
+    if(err) *perr = err_str;
     return err;
 }
 
@@ -556,6 +591,11 @@ bool translate_conn (struct starter_conn *conn
 	    {
 		if(!permitreplace)
 		{
+		    if(kw->keyword.keydef->validity & kv_duplicateok) {
+			/* be quiet about duplicate, but do not override it */
+			break;
+		    }
+			
 		    *error = _tmp_err;
 
 		    snprintf(_tmp_err, sizeof(_tmp_err)
@@ -1017,6 +1057,10 @@ int init_load_conn(struct starter_config *cfg
     
     connerr = load_conn (cfg, conn, cfgp, sconn, TRUE, perr);
 		
+    if(connerr != 0) {
+	starter_log(LOG_LEVEL_INFO, "while loading '%s': %s\n",
+		    sconn->name, *perr);
+    }
     if(connerr == 0)
     {
 	conn->state = STATE_LOADED;
