@@ -61,6 +61,8 @@
 #include "ocf_cryptodev.h"
 #endif
 
+TAILQ_HEAD(req_queue, pluto_crypto_req_cont);
+
 struct pluto_crypto_worker {
     int   pcw_helpernum;
     pid_t pcw_pid;
@@ -70,12 +72,11 @@ struct pluto_crypto_worker {
     int   pcw_maxcritwork;  /* how many critical things can be queued */
     bool  pcw_dead;         /* worker is dead, waiting for reap */
     bool  pcw_reaped;       /* worker has been reaped, waiting for dealloc */
-    struct pluto_crypto_req_cont *pcw_cont;
+    struct req_queue pcw_active;
 };
 
-static struct pluto_crypto_req_cont *backlogqueue;
-static struct pluto_crypto_req_cont **backlogqueue_last = &backlogqueue;
-static int                           backlogqueue_len;
+static struct req_queue backlog;
+static int       backlogqueue_len=0;
 
 static void init_crypto_helper(struct pluto_crypto_worker *w, int n);
 static void cleanup_crypto_helper(struct pluto_crypto_worker *w, int status);
@@ -367,10 +368,8 @@ err_t send_crypto_helper_request(struct pluto_crypto_req *r
 
     if(cnt == 0 && r->pcr_pcim >= pcim_demand_crypto) {
 	/* it is very important. Put it all on a queue for later */
-	if(!backlogqueue_last) backlogqueue_last = &backlogqueue;
-	*backlogqueue_last = cn;
-	cn->pcrc_next = NULL;
-	backlogqueue_last  = &cn->pcrc_next;
+	
+	TAILQ_INSERT_TAIL(&backlog, cn, pcrc_list);
 
 	/* copy the request */
 	r = clone_bytes(r, r->pcr_len, "saved cryptorequest");
@@ -405,8 +404,7 @@ err_t send_crypto_helper_request(struct pluto_crypto_req *r
     }
 
     /* link it to the active worker list */
-    cn->pcrc_next = w->pcw_cont;
-    w->pcw_cont = cn;
+    TAILQ_INSERT_TAIL(&w->pcw_active, cn, pcrc_list);
 
     passert(w->pcw_pid != -1);
     passert(w->pcw_pipe != -1);
@@ -432,10 +430,11 @@ static void crypto_send_backlog(struct pluto_crypto_worker *w)
     struct pluto_crypto_req_cont *cn;
 
     if(backlogqueue_len > 0) {
-	passert(backlogqueue != NULL);
+
+	passert(backlog.tqh_first != NULL);
+	cn = backlog.tqh_first;
+	TAILQ_REMOVE(&backlog, cn, pcrc_list);
 	
-	cn = backlogqueue;
-	backlogqueue = cn->pcrc_next;
 	backlogqueue_len--;
 	
 	r = cn->pcrc_pcr;
@@ -457,8 +456,7 @@ static void crypto_send_backlog(struct pluto_crypto_worker *w)
 	}
 	
 	/* link it to the active worker list */
-	cn->pcrc_next = w->pcw_cont;
-	w->pcw_cont = cn;
+	TAILQ_INSERT_TAIL(&w->pcw_active, cn, pcrc_list);
 	
 	passert(w->pcw_pid != -1);
 	passert(w->pcw_pipe != -1);
@@ -505,15 +503,12 @@ void delete_cryptographic_continuation(struct state *st)
 
     for(i=0; i<pc_workers_cnt; i++) {
 	struct pluto_crypto_worker *w = &pc_workers[i];
-	struct pluto_crypto_req_cont *cn, **cnp;
+	struct pluto_crypto_req_cont *cn;
 
-	cn = w->pcw_cont;
-	cnp = &w->pcw_cont;
-	while(cn && st->st_serialno != cn->pcrc_serialno) {
-	    cnp = &cn->pcrc_next;
-	    cn = cn->pcrc_next;
-	}
-	
+	for(cn = w->pcw_active.tqh_first;
+	    cn!=NULL && st->st_serialno != cn->pcrc_serialno;
+	    cn = cn->pcrc_list.tqe_next);
+		
 	if(cn == NULL) {
 	    DBG(DBG_CRYPT, DBG_log("no suspended cryptographic state for %lu\n"
 				   , st->st_serialno));
@@ -521,8 +516,7 @@ void delete_cryptographic_continuation(struct state *st)
 	}
 
 	/* unlink it, and free it */
-	*cnp = cn->pcrc_next;
-	cn->pcrc_next = NULL;
+	TAILQ_REMOVE(&w->pcw_active, cn, pcrc_list);
  
 	if(cn->pcrc_free) {
 	    /*
@@ -553,7 +547,7 @@ void handle_helper_comm(struct pluto_crypto_worker *w)
     struct pluto_crypto_req *r;
     int restlen;
     int actlen;
-    struct pluto_crypto_req_cont *cn, **cnp;
+    struct pluto_crypto_req_cont *cn;
 
     DBG(DBG_CRYPT|DBG_CONTROL
 	, DBG_log("helper %u has finished work (cnt now %d)"
@@ -627,13 +621,10 @@ void handle_helper_comm(struct pluto_crypto_worker *w)
     crypto_send_backlog(w);
 
     /* now match up request to continuation, and invoke it */
-    cn = w->pcw_cont;
-    cnp = &w->pcw_cont;
-    while(cn && r->pcr_id != cn->pcrc_id) {
-	cnp = &cn->pcrc_next;
-	cn = cn->pcrc_next;
-    }
-
+    for(cn = w->pcw_active.tqh_first;
+	cn!=NULL && r->pcr_id != cn->pcrc_id;
+	cn = cn->pcrc_list.tqe_next);
+		
     if(cn == NULL) {
 	loglog(RC_LOG_SERIOUS
 	       , "failed to find continuation associated with req %u\n",
@@ -642,8 +633,7 @@ void handle_helper_comm(struct pluto_crypto_worker *w)
     }
 
     /* unlink it */
-    *cnp = cn->pcrc_next;
-    cn->pcrc_next = NULL;
+    TAILQ_REMOVE(&w->pcw_active, cn, pcrc_list);
  
     passert(cn->pcrc_func != NULL);
 
@@ -682,6 +672,7 @@ static void init_crypto_helper(struct pluto_crypto_worker *w, int n)
     w->pcw_work     = 0;
     w->pcw_reaped = FALSE;
     w->pcw_dead   = FALSE;
+    TAILQ_INIT(&w->pcw_active);
 
     /* set the send/received queue length to be at least maxcritwork
      * times sizeof(pluto_crypto_req) in size
@@ -815,6 +806,8 @@ void init_crypto_helpers(int nhelpers)
     pc_workers_cnt = 0;
     pcw_id = 1;
 
+    TAILQ_INIT(&backlog);
+
     /* find out how many CPUs there are, if nhelpers is -1 */
     /* if nhelpers == 0, then we do all the work ourselves */
     if(nhelpers == -1) {
